/* tadd: embedding CPP tracers in C programs.
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tadd.l,v 1.14 2006/11/01 11:36:57 gondow Exp gondow $ */

%{
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "tadd.h"
#include "tadd.tab.h"
#include "misc.h"
#include "list.h"

#define UNUSED __attribute__((unused))

/* YY_USER_ACTION is always executed prior to the matched rule's action */
#define YY_USER_ACTION { yylloc.first_line   = yylloc.last_line;        \
                         yylloc.first_column = yylloc.last_column;      \
                         yylloc.last_column += yyleng;                  \
                         yylval.str = yylval.str2 = yylval.tadd = NULL; \
                         yylval.list = NULL;  \
                         add_str_to_directive (yytext); \
		       }
#define NEWLINE_UPDATE ( yylloc.last_line++, yylloc.last_column = 0 )
#define NEWLINE_DIRECTIVE ( yylval.offset = directive_size )
#define STRDUP         ( yylval.str  = estrdup (yytext),  \
                         yylval.str2 = estrdup (yytext),  \
                         yylval.tadd = estrdup (yytext) )

#define PP_TOKEN_TEXT(type, text) \
                       ( fprintf (lexout, "<pp_token " LINE_FMT     \
				  "type='" type "'>%s</pp_token>\n", \
				  LINE_INFO(yylloc), text)  )
#define PP_TOKEN_XML(type, text) \
                       do { char *xml = str_to_xml (text);           \
                         fprintf (lexout, "<pp_token " LINE_FMT     \
				  "type='" type "'>%s</pp_token>\n", \
				  LINE_INFO(yylloc), xml);           \
			 EFREE (xml); } while (0)

#define WHITE_SPACE_XML(text)   \
                       do { char *xml = str_to_xml (text);           \
                         fprintf (lexout, "<white_space " LINE_FMT  \
				  ">%s</white_space>\n",             \
				  LINE_INFO(yylloc), xml);           \
			 EFREE (xml); } while (0)
int is_if_or_elif = 0;
int func_mc_nest = 0;

static char *string_buf  = NULL;
static int  string_size  = 0;
static int  string_max   = 0;
static char *comment_buf = NULL;
static int  comment_size = 0;
static int  comment_max  = 0;

char *directive_buf = NULL;
static int  directive_max  = 0;
int  directive_size = 0;

static int comment_first_line, comment_first_column;
static int string_first_line,  string_first_column;

static void add_char_to_string  (char c);
static void add_char_to_comment (char c);
static void add_str_to_directive (char *s);
static char *delete_comment (char *s) UNUSED;
%}

%x comment string include define_head
%s if_or_elif

 /* abbreviation definitions ('*' means 'optional') 
      D : decimal digit,   O : octal digit,      H : hexadecimal digit
      L : letter,          E : signed exponent
      IS: integer suffix*, FS: floating suffix*
      WS: whitespace without newline */

O      [0-7]
D      [0-9]
H      [a-fA-F0-9]
IS     (([uU][lL]?)|([uU](ll|LL))|([lL][uU]?)|((ll|LL)[uU]?))
L      [a-zA-Z_$]
FS     [fFlL]?
E      [Ee][+-]?{D}+
WS     [ \t\f\v\r]           

L_INTEGER_CONSTANT  (((0{O}+)|({D}+)|(0[xX]{H}+)){IS}?)
L_FLOATING_CONSTANT (({D}+{E}{FS})|({D}*"."{D}+{E}?{FS})|({D}+"."{D}*{E}?{FS}))

L_SIMPLE_ESCAPE_SEQUENCE \\[ntvbrfa\\\?'"]
L_OCTAL_ESCAPE_SEQUENCE  \\{O}{1,3}
L_HEXADECIMAL_SEQUENCE   \\(x|X){H}+
L_UNIVERSAL_CHARACTER_NAME \\(u{H}{4}|U{H}{8})
L_ESCAPE_SEQUENCE        {L_SIMPLE_ESCAPE_SEQUENCE}|{L_OCTAL_ESCAPE_SEQUENCE}|{L_HEXADECIMAL_SEQUENCE}|{L_UNIVERSAL_CHARACTER_NAME}
L_CHARACTER_CONSTANT     (L?'([^\\\n']|{L_ESCAPE_SEQUENCE})+')
 /*
L_KEYWORD    auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|int|long|register|return|short|signed|sizeof|static|struct|switch|typedef|union|unsigned|void|volatile|while|inline|restrict|_Bool|_Complex|_Imaginary
 */
L_IDENTIFIER ({L}|{L_UNIVERSAL_CHARACTER_NAME})({L}|{D})*
L_CONSTANT {L_INTEGER_CONSTANT}|{L_FLOATING_CONSTANT}|{L_CHARACTER_CONSTANT}
L_PUNCTUATOR      "["|"]"|"{"|"}"|"."|"->"|"++"|"--"|"&"|"*"|"+"|"-"|"~"|"!"|"/"|"%"|"<<"|">>"|"<"|">"|"<="|">="|"=="|"!="|"^"|"|"|"&&"|"||"|"?"|":"|";"|"..."|"="|"*="|"/="|"%="|"+="|"-="|"<<="|">>="|"&="|"^="|"|="|","|"#"|"##"|"<:"|":>"|"<%"|"%>"|"%:"|"%:%:"

%% /* ==================================================================== */

^{WS}*#{WS}*include  { BEGIN(include);
                       yylval.offset = directive_size;
                       yylval.str  = estrdup (yytext);
                       PP_TOKEN_XML ("directive", yytext);
		       return INCLUDE; }
^{WS}*#{WS}*include_next  { BEGIN(include);
                       yylval.offset = directive_size;
                       yylval.str  = estrdup (yytext);
                       PP_TOKEN_XML ("directive", yytext);
		       return INCLUDE_NEXT; }
<include>{WS}+       { WHITE_SPACE_XML(yytext); }
<include>\<[^\n>]+\> { STRDUP; PP_TOKEN_XML ("header_name", yytext);
                       return H_HEADER_NAME; }
<include>\"[^\n"]+\" { STRDUP; PP_TOKEN_XML ("header_name", yytext);
                       return Q_HEADER_NAME; }
<include>.           { BEGIN(INITIAL); unput(yytext[0]);
                       yylloc.first_column--; yylloc.last_column--;
                       directive_size--; }
<include>\n    	     { BEGIN(INITIAL); NEWLINE_UPDATE; NEWLINE_DIRECTIVE;
                       WHITE_SPACE_XML(yytext); return NEWLINE;}
^{WS}*#{WS}*define{WS}* { BEGIN(define_head);
                       yylval.offset = directive_size;
                       yylval.str  = estrdup (yytext); 
                       PP_TOKEN_XML ("directive", yytext);
                       return DEFINE; }
<define_head>{WS}+   { WHITE_SPACE_XML(yytext); }
<define_head>{L_IDENTIFIER}       {
                       BEGIN(INITIAL);
                       yylval.str  = estrdup (yytext);
                       easprintf (&yylval.str2, "%s%s",
                                  estrdup (yytext), *double_suffix_p);
	               yylval.tadd = estrdup (yylval.str2);
                       yylval.offset = directive_size;
                       PP_TOKEN_TEXT ("identifier", yytext);
                       return IDENTIFIER; }
<define_head>{L_IDENTIFIER}"("    { 
                       char *str2;
                       BEGIN(INITIAL);
                       yylval.str  = estrdup (yytext);
		       str2 = estrdup (yytext);
		       chop (str2);
                       easprintf (&yylval.str2, "%s%s(",
						str2, *double_suffix_p);
		       EFREE (str2);
	               yylval.tadd = estrdup (yylval.str2);
                       yylval.offset = directive_size;

                       fprintf (lexout, "<pp_token " LINE_FMT
                                "type='identifier'>%s</pp_token>\n", idnum,
                                yylloc.first_line, yylloc.first_column,
                                yylloc.last_line, yylloc.last_column - 1,
                                chop (yytext));
                       fprintf (lexout, "<pp_token " LINE_FMT
                                "type='punctuator'>%s</pp_token>\n", idnum,
                                yylloc.last_line, yylloc.last_column - 1,
                                yylloc.last_line, yylloc.last_column,
                                "(");
                       return IDENTIFIER2;
                     }


^{WS}*#{WS}*undef{WS}* { yylval.str  = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return UNDEF; }
^{WS}*#{WS}*if       { BEGIN(if_or_elif); yylval.str = estrdup (yytext);
                       yylval.offset = directive_size; is_if_or_elif = 1;
                       PP_TOKEN_XML ("directive", yytext); return IF; }
^{WS}*#{WS}*elif     { BEGIN(if_or_elif); yylval.str = estrdup (yytext);
                       yylval.offset = directive_size; is_if_or_elif = 1;
                       PP_TOKEN_XML ("directive", yytext); return ELIF;}
<if_or_elif>\n       { BEGIN(INITIAL); is_if_or_elif = 0;
                       NEWLINE_UPDATE; NEWLINE_DIRECTIVE;
		       WHITE_SPACE_XML(yytext); return NEWLINE;}
^{WS}*#{WS}*ifdef    { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return IFDEF; }
^{WS}*#{WS}*ifndef   { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return IFNDEF; }
^{WS}*#{WS}*else     { yylval.str = estrdup (yytext); 
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return ELSE; }
^{WS}*#{WS}*endif    { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return ENDIF; }
^{WS}*#{WS}*pragma   { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return PRAGMA; }
^{WS}*#{WS}*line     { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return LINE; }
^{WS}*#{WS}*error    { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return ERROR; }
^{WS}*#{WS}*         { yylval.str = estrdup (yytext);
                       yylval.offset = directive_size;
                       PP_TOKEN_XML ("directive", yytext); return EMPTY; }

"("                  { yylval.offset = directive_size;
                       PP_TOKEN_TEXT ("punctuator", yytext);
                       return '('; }
")"                  { yylval.offset = directive_size;
                       PP_TOKEN_TEXT ("punctuator", yytext);
                       return ')'; }
","                  { PP_TOKEN_TEXT ("punctuator", yytext); return ','; }
"..."                { PP_TOKEN_TEXT ("punctuator", yytext); return ELLIPSIS; }
"#"                  { PP_TOKEN_TEXT ("punctuator", yytext); return '#'; }
"##"                 { PP_TOKEN_TEXT ("punctuator", yytext);
                       return SHARPSHARP; }
 /* even keyword can be replaced when CPP is working, so comment out is a must
{L_KEYWORD}          { STRDUP; return KEYWORD; }
 */
{L_IDENTIFIER}       { yylval.str  = estrdup (yytext);
                       easprintf (&yylval.str2, "%s%s",
                                  estrdup (yytext), *double_suffix_p);
	               yylval.tadd = estrdup (yylval.str2);
                       yylval.offset = directive_size;
                       PP_TOKEN_TEXT ("identifier", yytext);
                       return IDENTIFIER; }
{L_CONSTANT}         { STRDUP; 
                       if (yytext [0] == 'L' || yytext [0] == '\'') {
			   PP_TOKEN_XML ("char_constant", yytext);
		       } else {
			   PP_TOKEN_TEXT("pp_number", yytext);
		       }
                       return CONSTANT; }

 /* string-literal */
L?\"    	     { BEGIN(string); 
                       add_char_to_string ('"'); 
                       string_first_line   = yylloc.first_line;
                       string_first_column = yylloc.first_column; }
<string>"\\"[\\\"]   { add_char_to_string (yytext [0]); 
                       add_char_to_string (yytext [1]); }
<string>"\""   	     { char *xml;
                       BEGIN(INITIAL);
                       add_char_to_string ('\"'); 
                       yylval.str  = estrdup (string_buf);
                       yylval.str2 = estrdup (string_buf);
                       yylval.tadd = estrdup (string_buf);

		       xml = str_to_xml (string_buf);
                       fprintf (lexout, "<pp_token " LINE_FMT
                                "type='string_literal'>%s</pp_token>\n", idnum,
                                string_first_line, string_first_column,
                                yylloc.last_line, yylloc.last_column, xml);
                       EFREE (xml); EFREE (string_buf); string_buf = NULL;
		       return STRING_LITERAL; }
<string>"\n"   	     { // admitting illegal string literals
                       add_char_to_string ('\n'); 
                       NEWLINE_UPDATE; }
<string>.    	     { add_char_to_string (yytext [0]); }

{L_PUNCTUATOR}       { STRDUP; PP_TOKEN_XML ("punctuator", yytext);
                       return PUNCTUATOR; }
"\\\n"|"\\\r\n"      { NEWLINE_UPDATE;
		       WHITE_SPACE_XML(yytext); }  /* continuation line */
"\n"                 { NEWLINE_UPDATE;
                       NEWLINE_DIRECTIVE;
		       WHITE_SPACE_XML(yytext);
                       return NEWLINE; }
{WS}+                { WHITE_SPACE_XML(yytext); }
 
 /* comment */
"/*"	    	     { BEGIN(comment);
                       add_char_to_comment ('/');
                       add_char_to_comment ('*');
                       comment_first_line   = yylloc.first_line;
                       comment_first_column = yylloc.first_column; }
<comment>"*/" 	     { char *xml;
                       BEGIN(INITIAL);
                       add_char_to_comment ('*'); add_char_to_comment ('/');
		       xml = str_to_xml (comment_buf);
                       fprintf (lexout, "<comment " LINE_FMT
                                "type='delimited'>%s</comment>\n", idnum,
                                comment_first_line, comment_first_column,
                                yylloc.last_line, yylloc.last_column, xml);
                       EFREE (xml); EFREE (comment_buf); comment_buf = NULL; }
<comment>\n          { add_char_to_comment ('\n'); NEWLINE_UPDATE; }
<comment>.    	     { add_char_to_comment (yytext [0]); }
"//".*               { char *xml = str_to_xml (yytext);
                       fprintf (lexout, "<comment " LINE_FMT
                                "type=\"single_line\">%s</comment>\n", idnum,
                                yylloc.first_line, yylloc.first_column,
                                yylloc.last_line,  yylloc.last_column, xml); 
		       EFREE (xml); }
.                    {    STRDUP;
                          if (strchr (yytext, '@') != NULL)
                              fprintf (stderr, "!!!! '@' appeared (%d.%d)\n",
                              yylloc.first_line, yylloc.first_column);
			  PP_TOKEN_XML ("non_white_space_char", yytext);
                          return NON_WHITE_SPACE_CHAR;
                     }
%% /* ==================================================================== */
static void
add_char_to_string (char c)
{
    /* initialized */
    if (string_buf == NULL) {
	string_max = 16;
	string_buf = erealloc (string_buf, string_max);
	string_buf [0] = '\0';
	string_size = 1;
    }

    /* allocated enough? */
    if (string_max < string_size + 1) {
	string_max *= 2;
	string_buf = erealloc (string_buf, string_max);
    }
    string_buf [string_size - 1] = c;
    string_buf [string_size]     = '\0';
    string_size++;
}

static void
add_char_to_comment (char c)
{
    /* initialized? */
    if (comment_buf == NULL) {
	comment_max = 16;
	comment_buf = erealloc (comment_buf, comment_max);
	comment_buf [0] = '\0';
	comment_size = 1;
    }

    /* allocated enough? */
    if (comment_max < comment_size + 1) {
	comment_max *= 2;
	comment_buf = erealloc (comment_buf, comment_max);
    }
    comment_buf [comment_size - 1] = c;
    comment_buf [comment_size]     = '\0';
    comment_size++;
}

static void
add_str_to_directive (char *s)
{
    int len = strlen (s);

    /* initialized? */
    if (directive_buf == NULL) {
	directive_max = 16;
	directive_buf = erealloc (directive_buf, directive_max);
	directive_buf [0] = '\0';
	directive_size = 1;
    }

    /* allocated enough? */
    while (directive_max < directive_size + len) {
	directive_max *= 2;
	directive_buf = erealloc (directive_buf, directive_max);
    }
    memcpy (directive_buf + directive_size - 1, s, len + 1);
    directive_size += len;
}

static char *
delete_comment (char *s)
{
    char *p = estrdup (s), *q = p;
    
    while (*s != '\0') {
	if (s [0] == '/' && s [1] == '*') {
	    while (1) {
		if (s [0] == '\0')
		    break;
		if (s [0] == '*' && s [1] == '/') {
		    s += 2;
		    break;
		}
		s++;
	    }
	}
	if (s [0] == '/' && s [1] == '/') {
	    while (1) {
		if (s [0] == '\0')
		    break;
		if (s [0] == '\n') {
		    s++;
		    break;
		}
		s++;
	    }
	}
	*p++ = *s++;
    }
    *p = '\0';
    return q;
}
/* ==================================================================== */
void
tadd_unput (char *s)
{
    char *p = s + strlen (s) - 1;
    while (p >= s) {
	unput (*p--);
    }
}

#if 0
int
main (void)
{
    int token;
    while (token = yylex ()) {
	fprintf (stderr, "token = %d, lexeme = |%s|\n", token, yytext);
    }
}
#endif

