/* tadd: embedding CPP tracers into C programs.
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tadd.y,v 1.48 2006/11/08 13:37:40 gondow Exp gondow $ */

%{
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/param.h>
#include <signal.h>

#include "tadd.h"
#include "misc.h"
#include "pi.h"
#include "list.h"
#include "tef.h"

#define YYDEBUG 1
#define YYERROR_VERBOSE

#define NULLIFY(y)  ((y).str = (y).str2 = (y).tadd = NULL, (y).list = NULL)
#define MOVE(x,y)   ((x).str  = (y).str,  (x).str2 = (y).str2, \
                     (x).tadd = (y).tadd, (x).list = (y).list, \
                     NULLIFY(y))
#define MOVE1_FMT(x,y,fmt)    (easprintf (&x.str,  fmt, y.str),   \
			       easprintf (&x.str2, fmt, y.str2),  \
			       easprintf (&x.tadd, fmt, y.tadd),  \
			       x.list = NULL, free_lval (&y))
#define MOVE2_FMT(x,y,z,fmt)  (easprintf (&x.str,  fmt, y.str,  z.str),   \
			       easprintf (&x.str2, fmt, y.str2, z.str2),  \
			       easprintf (&x.tadd, fmt, y.tadd, z.tadd),  \
			       x.list = NULL, free_lval (&y), free_lval (&z))
#define BOTTOM(x,y) (x.str  = estrdup (y), x.str2 = estrdup (y), \
                     x.tadd = estrdup (y), x.list = NULL)

FILE *lexout;
char *absolute_path;
int  idnum;

static void lex_sync (void);
static void free_lval  (struct lval *lval);
static void free_lval2 (struct lval *lval);
static int  yyerror (char *mesg);
static void print_usage (void);
static void process_options (int argc, char *argv []);
static void output_xml_begin (char *file);
static void output_xml_end   (void);
static void output_params    (struct CDLL *list);
static void output_params2   (struct CDLL *list);
static char *get_text (int from, int to);
static void output_check_def_macro (char *macro_name, int is_defined);
static void tadd (char *file);
static void output_error_text (int idnum, int first_line, int first_column,
			       int last_line, int last_column);

static struct CDLL *cdll_head = NULL;
static int id_counter         = 0;
static int define_func_flag   = 0;
static struct {
    char **file;
    int  file_num;
    char *dir;
    char *db_file;
    char *lex_suffix;
    char *double_suffix;
    char *check_def_suffix;
    int  conservative;
} args = {
    NULL,          /* char **file; */
    0,             /* int  file_num; */
    "./TBCppA",    /* char *dir; */
    "db_file",     /* char *db_file; */
    ".lex",        /* char *lex_suffix; */
    "_1528552",    /* char *double_suffix; */ /* Titech zip code :-) */
    "_3141592_is_defined", /* char *check_def_suffix; */
    0,             /* int  conservative; */
};

char **double_suffix_p = &args.double_suffix;

%}

%error-verbose
%expect    2
%expect-rr 0
%start group

%token  IDENTIFIER IDENTIFIER2 CONSTANT STRING_LITERAL PUNCTUATOR NEWLINE NON_WHITE_SPACE_CHAR H_HEADER_NAME Q_HEADER_NAME

%token INCLUDE      "#include"
%token INCLUDE_NEXT "#include_next"
%token DEFINE       "#define"
%token UNDEF        "#undef"
%token IF           "#if"
%token IFDEF        "#ifdef"
%token IFNDEF       "#ifndef"
%token ELIF         "#elif"
%token ELSE         "#else"
%token ENDIF        "#endif"
%token ELLIPSIS     "..."
%token SHARPSHARP   "##"
%token LINE         "#line"
%token ERROR        "#error"
%token PRAGMA       "#pragma"
%token EMPTY        "#empty"

%left  SHARPSHARP  /* C99 says the eval order of ## op is unspecified */
%nonassoc '#'

%% /* ==================================================================== */

DF_on:   { lex_sync (); define_func_flag = 1; cdll_head = NULL; } ;
DF_off:  { lex_sync (); define_func_flag = 0; cdll_head = NULL; } ;

group
    : /* empty */      { NULLIFY ($$); }
    | group group_part { NULLIFY ($$); }
    ;

group_part
    : {
	  fprintf (yyout, "@\"cond id='F%d_%d'\"\n", idnum, id_counter);
	  id_counter++;
      } 
      if_group elif_groups else_group "#endif" NEWLINE
      {
	  char *text = get_text ($5.offset, $6.offset);
	  char *xml  = str_to_xml2 (text);

	  fprintf (yyout, "%s%s\n", $5.str, text);
	  fprintf (yyout, "%s%s%s", $2.tadd, $3.tadd, $4.tadd);
	  fprintf (yyout, "@\"endif id='F%d_%d' dir='%s%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $5.str, xml, LINE_INFO(@5));
	  fprintf (yyout, "@\"/cond\"\n");

	  id_counter++;
	  free_lval (&$2); free_lval (&$3); free_lval (&$4); 
	  EFREE ($5.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include" H_HEADER_NAME NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml = str_to_xml2 (text);
	  fprintf (yyout, "@\"include id='F%d_%d' dir='%s' type='system' "
		   "file='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include" Q_HEADER_NAME NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml = str_to_xml2 (text);
	  fprintf (yyout, "@\"include id='F%d_%d' dir='%s' type='user' "
		   "file='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include" pp_tokens_define NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"include id='F%d_%d' dir='%s' type='pp_tokens' "
		   "file='%s' " LINE_FMT "\"\n"
		   "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$),
		   $2.str, $2.tadd);
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include_next" H_HEADER_NAME NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"include_next id='F%d_%d' dir='%s' type='system' "
		   "file='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include_next\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include_next" Q_HEADER_NAME NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"include_next id='F%d_%d' dir='%s' type='user' "
		   "file='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include_next\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#include_next" pp_tokens_define  NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout,
		   "@\"include_next id='F%d_%d' dir='%s' type='pp_tokens' "
		   "file='%s' " LINE_FMT "\"\n"
		   "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$),
		   $2.str, $2.tadd);
	  fprintf (yyout, "%s%s\n", $1.str, text);
	  fprintf (yyout, "@\"/include_next\"\n");

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#define" DF_on IDENTIFIER pp_tokens_define_opt DF_off NEWLINE
      {
	  char *text = get_text ($3.offset, $6.offset);
	  char *xml  = str_to_xml2 (text);

	  fprintf (yyout, "@\"define type='obj' id='F%d_%d' dir='%s' "
		   "name='%s' macro_body='%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, $3.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s%s\n", $1.str, $3.str, text); 
	  fprintf (yyout, "#ifndef %s\n", $3.str2);
	  fprintf (yyout, "#define %s @\"macro_body ref='F%d_%d'\"%s@\"/macro_body\"\n",
		   $3.str2, idnum, id_counter, $4.tadd); 
	  fprintf (yyout, "#endif\n");
	  output_check_def_macro ($3.str, 1);
	  
	  id_counter++; free_lval (&$3); free_lval (&$4);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#define" DF_on IDENTIFIER2 identifier_list_opt ')'
                pp_tokens_define_opt DF_off NEWLINE 
      { 
	  char *text, *text2, *xml, *xml2;
	  char *id2 = estrdup ($3.str);
	  char *double2 = estrdup ($3.str2);
	  chop (id2); chop (double2);

	  text = get_text ($5.offset, $8.offset);
	  xml  = str_to_xml2 (text);

	  text2 = get_text ($3.offset, $5.offset);
	  chop_char (text2, ')');
	  xml2 = str_to_xml2 (text2);

	  fprintf (yyout, "@\"define type='func' id='F%d_%d' dir='%s' "
		   "name='%s' args='%s' macro_body='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, id2, xml2, xml, LINE_INFO(@$));
	  output_params ($4.list);
	  fprintf (yyout, "@\"/define\"\n");
          fprintf (yyout, "%s%s%s)%s\n",
		   $1.str, $3.str, text2, text);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  fprintf (yyout, "#define %s%s%s%s) @\"macro_body ref='F%d_%d'\"",
		   $3.str2, $4.str, ((!strcmp ($4.str, "")) ? "" : ", "),
		   $4.str2, idnum, id_counter);
	  output_params2 ($4.list);
	  fprintf (yyout, "%s@\"/macro_body\"\n", $6.tadd);
	  fprintf (yyout, "#endif\n");
	  output_check_def_macro (id2, 1);

          id_counter++; free_lval (&$3); free_lval2 (&$4); free_lval (&$6);
	  EFREE ($1.str); EFREE (id2); EFREE (double2);
	  EFREE (text); EFREE (xml); EFREE (text2); EFREE (xml2);
	  NULLIFY($$);
      }
    | "#define" DF_on IDENTIFIER2 "..." ')'
                pp_tokens_define_opt DF_off NEWLINE 
      { 
	  char *text, *text2, *xml, *xml2;
	  char *id2 = estrdup ($3.str);
	  char *double2 = estrdup ($3.str2);
	  chop (id2); chop (double2);

	  text = get_text ($5.offset, $8.offset);
	  xml = str_to_xml2 (text);

	  text2 = get_text ($3.offset, $5.offset);
	  chop_char (text2, ')');
	  xml2 = str_to_xml2 (text2);

	  fprintf (yyout, "@\"define type='func_variadic' id='F%d_%d' "
		   "dir='%s' name='%s' args='%s' macro_body='%s' "
		   LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, id2, xml2, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s%s) %s\n",
		   $1.str, $3.str, text2, text);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  fprintf (yyout, "#define %s...) @\"macro_body ref='F%d_%d'\""
		   "@\"param name='__VA_ARGS__'\"(__VA_ARGS__)@\"/param\""
		   "TBCPPA_APPLY(%s, TBCPPA_FIRST_HALF_ARGS(__VA_ARGS__))"
		   "@\"/macro_body\"\n",
		   $3.str2, idnum, id_counter, id2);
	  fprintf (yyout, "#endif\n");
	  output_check_def_macro (id2, 1);

	  id_counter++; free_lval (&$3); free_lval (&$6);
	  EFREE ($1.str); EFREE (id2); EFREE (double2);
	  EFREE (text); EFREE (xml); EFREE (text2); EFREE (xml2);
      }
    | "#define" DF_on IDENTIFIER2 identifier_list ',' "..." ')'
                pp_tokens_define_opt DF_off NEWLINE
      {
	  char *text, *text2, *xml, *xml2;
	  char *id2 = estrdup ($3.str);
	  char *double2 = estrdup ($3.str2);
	  chop (id2); chop (double2);

	  text = get_text ($7.offset, $10.offset);
	  xml  = str_to_xml2 (text);

	  text2 = get_text ($3.offset, $7.offset);
	  chop_char (text2, ')');
	  xml2 = str_to_xml2 (text2);

	  fprintf (yyout, "@\"define type='func_variadic' id='F%d_%d' "
		   "dir='%s' name='%s' args='%s' macro_body='%s' "
		   LINE_FMT "\"\n",
		   idnum, id_counter, $1.str, id2, xml2, xml, LINE_INFO(@$));
	  output_params ($4.list);
	  fprintf (yyout, "@\"/define\"\n");
	  fprintf (yyout, "%s%s%s) %s\n",
		   $1.str, $3.str,  text2, text);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  fprintf (yyout,
		   "#define %s%s,...) @\"macro_body ref='F%d_%d'\"",
		   $3.str2, $4.str, idnum, id_counter);
	  output_params2 ($4.list);
	  fprintf (yyout,
		   "@\"param name='__VA_ARGS__'\"(__VA_ARGS__)@\"/param\""
		   "TBCPPA_APPLY(%s, TBCPPA_FIRST_HALF_ARGS(%s, __VA_ARGS__))"
		   "@\"/macro_body\"\n", id2, $4.str);
	  fprintf (yyout, "#endif\n");
	  output_check_def_macro (id2, 1);

	  id_counter++; free_lval (&$3); free_lval2 (&$4); free_lval (&$8);
	  EFREE ($1.str); EFREE (id2); EFREE (double2);
	  EFREE (text); EFREE (xml); EFREE (text2); EFREE (xml2);
	  NULLIFY($$);
      }
    | "#undef"  IDENTIFIER NEWLINE
      {
	  char *text = get_text ($2.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);

	  fprintf (yyout, "@\"undef id='F%d_%d' dir='%s' name='%s%s' "
		   LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, $2.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s%s\n", $1.str, $2.str, text);
	  fprintf (yyout, "#undef %s\n", $2.str2);
	  output_check_def_macro ($2.str, 0);

	  id_counter++; free_lval (&$2);
	  EFREE ($1.str); EFREE (text); EFREE (xml);
	  NULLIFY($$);
      }
    | "#line"   pp_tokens_define     NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"other_cpp_directive type='line' id='F%d_%d' "
		   "dir='%s%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);

	  id_counter++; free_lval (&$2);
	  EFREE (text); EFREE (xml); EFREE ($1.str);
	  NULLIFY($$);
      }
    | "#error"  pp_tokens_define_opt NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"other_cpp_directive type='error' id='F%d_%d' "
		   "dir='%s%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);

	  id_counter++; free_lval (&$2);
	  EFREE (text); EFREE (xml); EFREE ($1.str);
	  NULLIFY($$);
      }
    | "#pragma" pp_tokens_define_opt NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"other_cpp_directive type='pragma' id='F%d_%d' "
		   "dir='%s%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);

	  id_counter++; free_lval (&$2);
	  EFREE (text); EFREE (xml); EFREE ($1.str);
	  NULLIFY($$);
      }
    | "#empty" pp_tokens_define_opt NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "@\"other_cpp_directive type='empty' id='F%d_%d' "
		   "dir='%s%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $1.str, xml, LINE_INFO(@$));
	  fprintf (yyout, "%s%s\n", $1.str, text);

	  id_counter++; free_lval (&$2);
	  EFREE (text); EFREE (xml); EFREE ($1.str);
	  NULLIFY($$);
      }
    | pp_tokens_opt NEWLINE
      {
	  fprintf (yyout, "%s\n", $1.tadd);
	  free_lval (&$1);
	  NULLIFY($$);
      }
    | error NEWLINE
      {
	  yyerrok;
	  fprintf (yyout, "/* error recovery (F%d:%d.%d-%d.%d) */\n",
		   LINE_INFO(@$));
	  output_error_text (LINE_INFO(@$));
	  fprintf (yyout, "/* end of error recovery */\n");
	  NULLIFY($$);
      }
    ;

if_group
    : "#if"     pp_tokens_define  NEWLINE 
      { 
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml = str_to_xml2 (text);
	  fprintf (yyout, "%s%s\n", $1.str, text);
          fprintf (yyout, "@\"cond_selected type='if' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $2.tadd);
	  $$.str  = (char *)id_counter++; 
	  $$.str2 = (char *)xml;
	  EFREE (text);
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "@\"if id='F%d_%d' dir='%s' exp='%s' " LINE_FMT "/\"\n",
		     idnum, (int)$4.str, $1.str, (char *)$4.str2,
		     LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
	  free_lval (&$2);
	  EFREE ($1.str);
	  EFREE ((char *)$4.str2);
      }
    | "#ifdef"  IDENTIFIER NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "%s%s\n",  $1.str, text);
          fprintf (yyout, "@\"cond_selected type='ifdef' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $2.tadd);
	  $$.str  = (char *)id_counter++; 
	  $$.str2 = (char *)xml;
	  EFREE (text);
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd, "@\"ifdef id='F%d_%d' dir='%s' exp='%s' "
		     LINE_FMT "/\"\n",
		     idnum, (int)$4.str, $1.str, (char *)$4.str2,
		     LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
	  EFREE ($1.str);
	  EFREE ((char *)$4.str2);
	  free_lval (&$2);
      }
    | "#ifndef" IDENTIFIER NEWLINE
      {
	  char *text = get_text ($1.offset, $3.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "%s%s\n", $1.str, text);
          fprintf (yyout, "@\"cond_selected type='ifndef' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1, @3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $2.tadd);
	  $$.str  = (char *)id_counter++; 
	  $$.str2 = (char *)xml;
	  EFREE (text);
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd, "@\"ifndef id='F%d_%d' dir='%s' exp='%s' "
		     LINE_FMT "/\"\n",
		     idnum, (int)$4.str, $1.str, (char *)$4.str2,
		     LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
	  EFREE ($1.str);
	  EFREE ((char *)$4.str2);
	  free_lval (&$2);
      }
    ;

elif_groups
    : /* empty */
      {
	  NULLIFY ($$);
	  $$.tadd = estrdup ("");
      }
    | elif_groups "#elif" pp_tokens_define NEWLINE
      {
	  char *text = get_text ($2.offset, $4.offset);
	  char *xml  = str_to_xml2 (text);
	  fprintf (yyout, "%s%s\n", $2.str, text);
          fprintf (yyout, "@\"cond_selected type='elif' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@2,@4));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $3.tadd);
	  $$.str  = (char *)id_counter++; 
	  $$.str2 = (char *)xml;
	  EFREE (text);
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "%s@\"elif id='F%d_%d' dir='%s' exp='%s' "
		     LINE_FMT "/\"\n",
		     $1.tadd, idnum, (int)$5.str, $2.str, (char *)$5.str2,
		     LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
	  free_lval (&$1); free_lval (&$3);
	  EFREE ($2.str); EFREE ((char *)$5.str2);
      }
    ;

else_group
    : /* empty */
      {
	  NULLIFY ($$);
	  $$.tadd = estrdup ("");
      }
    | "#else" NEWLINE
      {
	  char *text = get_text ($1.offset, $2.offset);
	  char *xml = str_to_xml2 (text);
	  fprintf (yyout, "%s%s\n", $1.str, text);
          fprintf (yyout, "@\"cond_selected type='else' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@2));
	  $$.str  = (char *)id_counter++; 
	  $$.str2 = (char *)xml;
	  EFREE (text);
      }
      group 
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd, "@\"else id='F%d_%d' dir='%s%s' " 
		     LINE_FMT "/\"\n",
		     idnum, (int)$3.str, $1.str, (char *)$3.str2,
		     LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
	  EFREE ($1.str); EFREE ((char *)$3.str2);
      }
    ;

identifier_list_opt
    : /* empty */
      {
	  $$.str  = estrdup ("");
	  $$.str2 = estrdup ("");
	  $$.tadd = NULL;
	  $$.list = emalloc (sizeof (struct CDLL));
	  CDLL_init ($$.list);
	  cdll_head = $$.list;
      }
    | identifier_list
      {
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;
	  $$.tadd = NULL;
	  $$.list = $1.list;
	  $1.list = NULL;
	  cdll_head = $$.list;
      }
    ;

identifier_list
    : IDENTIFIER
      {
	  $$.str  = estrdup ($1.str);
	  $$.str2 = estrdup ($1.str2);
	  $$.tadd = NULL;
	  $$.list = emalloc (sizeof (struct CDLL));
	  CDLL_init ($$.list); 
	  CDLL_insert ($$.list, estrdup ($1.str));
	  free_lval (&$1);
      }
    | identifier_list ',' IDENTIFIER
      {
	  easprintf (&$$.str,  "%s, %s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s, %s", $1.str2, $3.str2);
	  $$.tadd = NULL;
	  $$.list = $1.list;
	  CDLL_insert ($$.list, estrdup ($3.str));
	  free_lval (&$1); free_lval (&$3);
      }
    ;

identifier
    : IDENTIFIER
      {
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;

	  if (!strcmp ($1.str, "defined") && is_if_or_elif) {
	      /* '(' doesn't have to follow 'defined'  */
	      $$.tadd = estrdup ($1.str);
	  } else if (define_func_flag) {
	      if (CDLL_query (cdll_head, $1.str) != -1
		  || !strcmp ($1.str, "__VA_ARGS__")) {
		  /* a parameter in function-like macro */
		  easprintf (&$$.tadd, 
			     "@\"macro_call type='param' name='%s' "
			     LINE_FMT "\""
			     "@\"is_defined\"1@\"/is_defined\""
			     "@\"bare\"%s@\"/bare\" @\"double\"%s@\"/double\""
			     "@\"/macro_call\"",
			     $1.str, LINE_INFO(@1), $1.str,
			     ((!strcmp ($1.str, "__VA_ARGS__")) ?
			      $1.str: $1.str2));
	      } else {
		  easprintf (&$$.tadd, 
			     "@\"macro_call type='obj' name='%s' "
			     LINE_FMT "\""
			     "@\"is_defined\"%s%s@\"/is_defined\""
			     "@\"bare\"%s@\"/bare\" @\"double\"%s@\"/double\""
			     "@\"/macro_call\"",
			     $1.str, LINE_INFO(@1),
			     $1.str, args.check_def_suffix,
			     $1.str, $1.str2);
	      }
	  } else {
	      easprintf (&$$.tadd, 
			 "@\"macro_call type='obj' name='%s' " LINE_FMT "\""
			 "@\"is_defined\""
			 "\n#ifdef %s\n1\n#else\n0\n#endif\n"
			 "@\"/is_defined\""
			 "@\"bare\"%s@\"/bare\" @\"double\"%s@\"/double\""
			 "@\"/macro_call\"",
			 $1.str, LINE_INFO(@1),
			 $1.str, $1.str, $1.str2);
	  }

	  EFREE ($1.tadd);
	  NULLIFY ($1);
      }
    ;

/* pp_tokens for all CPP directives like #define */
pp_tokens_define_opt
    : /* empty */                { BOTTOM($$, ""); }
    | pp_tokens_define           { MOVE($$, $1);   }
    ;

pp_tokens_define
    : preprocessing_token_define { MOVE($$, $1);   }
    | pp_tokens_define preprocessing_token_define 
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_define
    : preprocessing_token_define_minus_paren_comma   { MOVE($$, $1); }
    | func_macro_call_define      { MOVE($$, $1); }
    | '('                         { BOTTOM($$, yytext);}
    | ')'                         { BOTTOM($$, yytext);}
    | ','                         { BOTTOM($$, yytext); }
    ;


preprocessing_token_define_minus_paren_comma
    : identifier                  { MOVE($$, $1); }
    | CONSTANT                    { MOVE($$, $1); }
    | STRING_LITERAL              { MOVE($$, $1); }
    | PUNCTUATOR                  { MOVE($$, $1); }
    | NON_WHITE_SPACE_CHAR        { MOVE($$, $1); }
    | "..."                       { BOTTOM($$, yytext);}
    | '#'  identifier
      {
	  /* eliminate tracers for #'s operands */
	  easprintf (&$$.str,  "#%s", $2.str);
	  easprintf (&$$.str2, "#%s", $2.str2);
	  easprintf (&$$.tadd, "#%s", $2.str); /* $2.str, not $2.tadd */
	  free_lval (&$2);
      }
    | preprocessing_token_define_minus_paren_comma
      "##" preprocessing_token_define_minus_paren_comma
      {
	  /* eliminate tracers for ##'s operands */
	  easprintf (&$$.str,  "%s##%s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s##%s", $1.str2, $3.str2);
	  easprintf (&$$.tadd, "%s##%s", $1.str,  $3.str); /* str, not tadd */
	  free_lval (&$1); free_lval (&$3);
      }
    ;

pp_tokens_opt
    : /* empty */           { BOTTOM($$, ""); }
    | pp_tokens             { MOVE($$, $1); }
    ;

pp_tokens
    : preprocessing_token            { MOVE($$, $1); }
    | pp_tokens preprocessing_token  { MOVE2_FMT($$, $1, $2, "%s %s"); }
    ;

preprocessing_token
    : preprocessing_token_minus_paren_comma   { MOVE($$, $1); }
    | func_macro_call_plus_nl                 { MOVE($$, $1); }
    | '('                                     { BOTTOM($$, yytext); }
    | ')'                                     { BOTTOM($$, yytext); }
    | ','                                     { BOTTOM($$, yytext); }
    ;

/* = pp_tokens_opt - ('(' + ')' + ',') */
preprocessing_token_minus_paren_comma
    : identifier            { MOVE($$, $1); }
    | CONSTANT              { MOVE($$, $1); }
    | STRING_LITERAL        { MOVE($$, $1); }
    | PUNCTUATOR            { MOVE($$, $1); }
    | NON_WHITE_SPACE_CHAR  { MOVE($$, $1); }
    | "..."                 { BOTTOM($$, yytext); }
    | '#'                   { BOTTOM($$, yytext); }
    | "##"                  { BOTTOM($$, yytext); }
    ;

func_mc_left_paren
    : '('
      {
	  assert (yychar == YYEMPTY);
	  func_mc_nest++;
	  $$.offset = $1.offset;
      }
    ;

func_mc_right_paren
    : ')'
      {
	  assert (yychar == YYEMPTY);
	  func_mc_nest--;
	  assert (func_mc_nest >= 0);
	  $$.offset = $1.offset;
      }
    ;

/* we don't have to take care of IDENTIFIER2 here, since IDENTIFIER2
   will appear only as #define's next identifier. */
func_macro_call_plus_nl
    : IDENTIFIER func_mc_left_paren argument_list func_mc_right_paren
      {
	  char *text  = get_text ($2.offset, $4.offset);
	  char *xml   = str_to_xml2 (chop_char (text, ')'));
	  char *text2 = get_text ($1.offset, $2.offset);
	  char *xml2  = str_to_xml2 (chop_char (text2, '('));

	  easprintf(&$$.str,  "%s%s(%s)", $1.str, text2, $3.str);
	  easprintf(&$$.str2, "%s(%s%s%s)", $1.str2, $3.str, 
		    ((!strcmp($3.str, "")) ? "" : ", "), $3.str2);

	  if (!strcmp ($1.str, "_Pragma")) {
	      $$.tadd = estrdup ($$.str);
	  } else if (!strcmp ($1.str, "defined") && is_if_or_elif) {
	      easprintf(&$$.tadd,
			"@\"macro_call type='defined' name='%s%s' args='%s' "
			LINE_FMT "\""
			"@\"is_defined\"1@\"/is_defined\""
			"@\"bare\"\n#ifdef %s\n1\n#else\n0\n#endif\n@\"/bare\""
			"@\"double\"%s@\"/double\""
			"@\"/macro_call\"",
			$1.str, xml2, xml, LINE_INFO(@$), $3.str, $3.tadd);
	  } else {
	      char *tadd1, *tadd2, *tadd3, *tadd4;

	      easprintf (&tadd1,
			 "@\"macro_call type='func' name='%s%s' args='%s' "
			 LINE_FMT "\"", $1.str, xml2, xml, LINE_INFO(@$));

	      if (define_func_flag) {
		  easprintf (&tadd2,
			     "@\"is_defined\"%s%s@\"/is_defined\"",
			     $1.str, args.check_def_suffix);
	      } else {
		  easprintf (&tadd2,
			     "@\"is_defined\""
			     "\n#ifdef %s\n1\n#else\n0\n#endif\n"
			     "@\"/is_defined\"",
			     $1.str);
	      }

	      easprintf (&tadd3, "@\"bare\"%s%s(%s)@\"/bare\"",
			 $1.str, text2, text);

	      if (args.conservative) {
		  easprintf (&tadd4,
			     "@\"double\"%s(%s)@\"/double\"", $1.str, $3.str2);
	      } else {
		  easprintf (&tadd4,
			     "@\"double\"%s(%s%s%s)@\"/double\"",
			     $1.str2, $3.str,
			     ((!strcmp($3.str, "")) ? "" : ", "), $3.tadd);
	      }

	      easprintf (&$$.tadd, "%s%s%s%s" "@\"/macro_call\"",
			 tadd1, tadd2, tadd3, tadd4);
	      EFREE (tadd1); EFREE (tadd2); EFREE (tadd3); EFREE (tadd4);
	  }

	  id_counter++; free_lval (&$1); free_lval (&$3);
	  EFREE (text); EFREE (xml); EFREE (text2); EFREE (xml2);
      }
    ;

func_macro_call_define
    : IDENTIFIER func_mc_left_paren argument_list_define func_mc_right_paren
      {
	  char *text  = get_text ($2.offset, $4.offset);
	  char *xml   = str_to_xml2 (chop_char (text, ')'));
	  char *text2 = get_text ($1.offset, $2.offset);
	  char *xml2  = str_to_xml2 (chop_char (text2, '('));

	  easprintf(&$$.str,  "%s%s(%s)", $1.str, text2, $3.str);
	  easprintf(&$$.str2, "%s(%s%s%s)", $1.str2, $3.str, 
		    ((!strcmp($3.str, "")) ? "" : ", "), $3.str2);

	  if (!strcmp ($1.str, "_Pragma")) {
	      $$.tadd = estrdup ($$.str);
	  } else if (!strcmp ($1.str, "defined") && is_if_or_elif) {
	      easprintf(&$$.tadd,
			"@\"macro_call type='defined' name='%s%s' args='%s' "
			LINE_FMT "\""
			"@\"is_defined\"1@\"/is_defined\""
			"@\"bare\"\n#ifdef %s\n1\n#else\n0\n#endif\n@\"/bare\""
			"@\"double\"%s@\"/double\""
			"@\"/macro_call\"",
			$1.str, xml2, xml, LINE_INFO(@$), $3.str, $3.tadd);
	  } else {
	      char *tadd1, *tadd2, *tadd3, *tadd4;

	      easprintf (&tadd1,
			 "@\"macro_call type='func' name='%s%s' args='%s' "
			 LINE_FMT "\"", $1.str, xml2, xml, LINE_INFO(@$));

	      if (define_func_flag) {
		  easprintf (&tadd2,
			     "@\"is_defined\"%s%s@\"/is_defined\"",
			     $1.str, args.check_def_suffix);
	      } else {
		  easprintf (&tadd2,
			     "@\"is_defined\""
			     "\n#ifdef %s\n1\n#else\n0\n#endif\n"
			     "@\"/is_defined\"",
			     $1.str);
	      }

	      easprintf (&tadd3, "@\"bare\"%s%s(%s)@\"/bare\"",
			 $1.str, text2, text);

	      if (args.conservative) {
		  easprintf (&tadd4,
			     "@\"double\"%s(%s)@\"/double\"", $1.str, $3.str2);
	      } else {
		  easprintf (&tadd4,
			     "@\"double\"%s(%s%s%s)@\"/double\"",
			     $1.str2, $3.str,
			     ((!strcmp($3.str, "")) ? "" : ", "), $3.tadd);
	      }

	      easprintf (&$$.tadd, "%s%s%s%s" "@\"/macro_call\"",
			 tadd1, tadd2, tadd3, tadd4);

	      EFREE (tadd1); EFREE (tadd2); EFREE (tadd3); EFREE (tadd4);
	  }

	  id_counter++; free_lval (&$1); free_lval (&$3);
	  EFREE (text); EFREE (xml); EFREE (text2); EFREE (xml2);
      }
    ;

pp_tokens_minus_paren_comma_plus_nl_opt
    : /* empty */                           { BOTTOM($$, ""); }
    | pp_tokens_minus_paren_comma_plus_nl   { MOVE($$, $1); }
    ;
    
pp_tokens_define_minus_paren_comma_plus_func_opt
    : /* empty */                           { BOTTOM($$, ""); }
    | pp_tokens_define_minus_paren_comma_plus_func    { MOVE($$, $1); }
    ;

pp_tokens_define_minus_paren_comma_plus_func
    : preprocessing_token_define_minus_paren_comma_plus_func { MOVE($$, $1); }
    | pp_tokens_define_minus_paren_comma_plus_func
      preprocessing_token_define_minus_paren_comma_plus_func
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;
    
pp_tokens_minus_paren_comma_plus_nl
    : preprocessing_token_minus_paren_comma_plus_nl  { MOVE($$, $1); }
    | pp_tokens_minus_paren_comma_plus_nl
      preprocessing_token_minus_paren_comma_plus_nl
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_minus_paren_comma_plus_nl
    : preprocessing_token_minus_paren_comma     { MOVE($$, $1); }
    | func_macro_call_plus_nl                   { MOVE($$, $1); }
    | '(' pp_tokens_minus_paren_plus_nl_opt ')' { MOVE1_FMT($$, $2, "(%s)"); }
    | NEWLINE                                   { BOTTOM($$, ""); }
    ;

preprocessing_token_define_minus_paren_comma_plus_func
    : preprocessing_token_define_minus_paren_comma  { MOVE($$, $1); }
    | func_macro_call_define                    { MOVE($$, $1); }
    | '(' pp_tokens_define_minus_paren_plus_func_opt ')'  { MOVE1_FMT($$, $2, "(%s)"); }
    ;

pp_tokens_define_minus_paren_plus_func_opt
    : /* empty */                             { BOTTOM($$, ""); }
    | pp_tokens_define_minus_paren_plus_func  { MOVE($$, $1); }
    ;

pp_tokens_define_minus_paren_plus_func
    : preprocessing_token_define_minus_paren_plus_func  { MOVE($$, $1); }
    | pp_tokens_define_minus_paren_plus_func
      preprocessing_token_define_minus_paren_plus_func
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_define_minus_paren_plus_func
    : preprocessing_token_define_minus_paren_comma_plus_func  { MOVE($$, $1); }
    | ','                                           { BOTTOM($$, yytext); }
    ;

pp_tokens_minus_paren_plus_nl_opt
    : /* empty */                     { BOTTOM($$, ""); }
    | pp_tokens_minus_paren_plus_nl   { MOVE($$, $1); }
    ;
    
pp_tokens_minus_paren_plus_nl
    : preprocessing_token_minus_paren_plus_nl { MOVE($$, $1); }
    | pp_tokens_minus_paren_plus_nl preprocessing_token_minus_paren_plus_nl
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_minus_paren_plus_nl
    : preprocessing_token_minus_paren_comma_plus_nl  { MOVE($$, $1); }
    | ','                                            { BOTTOM($$, yytext); }
    ;

argument_list
    : pp_tokens_minus_paren_comma_plus_nl_opt
      {
	  char *xml = str_to_xml2 ($1.str);
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "@\"arg exp='%s' " LINE_FMT "\"%s@\"/arg\"",
			 xml, LINE_INFO(@1), $1.tadd);
	      EFREE (xml);
	  }
	  EFREE ($1.tadd);
	  NULLIFY ($1);
      }
    | argument_list ',' pp_tokens_minus_paren_comma_plus_nl_opt
      {
	  char *xml = str_to_xml2 ($3.str);
	  easprintf (&$$.str,  "%s, %s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s, %s", $1.str2, $3.str2);
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "%s, @\"arg exp='%s' " LINE_FMT
			 "\"%s@\"/arg\"",
			 $1.tadd, xml, LINE_INFO(@3), $3.tadd);
	      EFREE (xml);
	  }
	  free_lval (&$1); free_lval (&$3);
	  NULLIFY ($1); NULLIFY ($3);
      }
    ;

argument_list_define
    : pp_tokens_define_minus_paren_comma_plus_func_opt
      {
	  char *xml = str_to_xml2 ($1.str);
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "@\"arg exp='%s' " LINE_FMT "\"%s@\"/arg\"",
			 xml, LINE_INFO(@1), $1.tadd);
	      EFREE (xml);
	  }
	  EFREE ($1.tadd);
	  NULLIFY ($1);
      }
    | argument_list_define ',' pp_tokens_define_minus_paren_comma_plus_func_opt
      {
	  char *xml = str_to_xml2 ($3.str);
	  easprintf (&$$.str,  "%s, %s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s, %s", $1.str2, $3.str2);
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "%s, @\"arg exp='%s' " LINE_FMT
			 "\"%s@\"/arg\"",
			 $1.tadd, xml, LINE_INFO(@3), $3.tadd);
	      EFREE (xml);
	  }
	  free_lval (&$1); free_lval (&$3);
	  NULLIFY ($1); NULLIFY ($3);
      }
    ;
%% /* ==================================================================== */
/* to make sure yyparse () does always fetch a look-ahead token eagerly. */
static void
lex_sync (void)
{
    if (yychar == YYEMPTY) {
        yychar = yylex ();
        /* just in case yylex returns negative to indicate EOF */
        if (yychar < 0) 
            yychar = 0; /* EOF */
    }
}

static void
free_lval (struct lval *lval)
{
    assert (lval != NULL);
    if (lval->str  != NULL) { EFREE (lval->str);  lval->str  = NULL; }
    if (lval->str2 != NULL) { EFREE (lval->str2); lval->str2 = NULL; }
    if (lval->tadd != NULL) { EFREE (lval->tadd); lval->tadd = NULL; }
}

static void
free_lval2 (struct lval *lval)
{
    assert (lval != NULL);
    free_lval (lval);
    if (lval->list != NULL) {
	CDLL_free (lval->list);
	lval->list = NULL;
    }
}

static int
yyerror (char *mesg)
{
#if 1 /* experimental error recovery */
    switch (yychar) {
    case INCLUDE: case INCLUDE_NEXT: case DEFINE: case UNDEF:
    case IF:      case IFDEF:        case IFNDEF: case ELIF:   case ELSE:
    case ENDIF:   case LINE:         case ERROR:  case PRAGMA: case EMPTY:
	tadd_unput (yytext);
	tadd_unput ("\n");
	yylloc.last_line--;
	yyclearin;
	break;
    default:
	/* do nothing */
	break;
    }
#endif

    return fprintf (stderr, "error(%d: %d.%d-%d.%d): %s\n",
		    LINE_INFO(yylloc), mesg);
}

static void
print_usage (void)
{
    fprintf (stderr, 
	     "usage  : tadd.exe [options] foo.[ch] ...\n"
	     "options:\n"
	     "    --dir         dir      output dir [./TBCppA]\n"
	     "    --db-file     file     db_flie [db_file](relative to dir)\n"
	     "    --lex-suffix suffix    suffix [.lex]\n"
	     "    --double-suffix suffix [_1528552]\n"
	     "    --check-def-suffix suffix [_3141592_is_defined]\n"
             "    --yydebug              enable bison debugging[off]\n"
	     "    --conservative         no embedding tracers to macro args\n"
	     "    --help                 prints this usage\n"
	     );
}

static void
process_options (int argc, char *argv [])
{
    int    i;
    for (i = 1; i < argc; i++) {
	if (argv [i][0] != '-') {                    /* filenames */
	    int j;
	    for (j = 0; i < argc && argv [i][0] != '-'; i++, j++) {
		args.file = erealloc (args.file, sizeof (char *) * (j + 1));
		args.file [j] = emalloc (strlen (argv [i]) + 1);
		strcpy (args.file [j], argv [i]);
	    }
	    i--;
	    args.file_num = j;
	} else if (!strcmp (argv [i], "--dir")) {
	    i++;
	    args.dir = argv [i];
	} else if (!strcmp (argv [i], "--db-file")) {
	    i++;
	    args.db_file = argv [i];
	} else if (!strcmp (argv [i], "--lex-suffix")) {
	    i++;
	    args.lex_suffix = argv [i];
	} else if (!strcmp (argv [i], "--yydebug")) {
	    yydebug = 1;
	} else if (!strcmp (argv [i], "--double-suffix")) {
	    i++;
	    args.double_suffix = argv [i];
	} else if (!strcmp (argv [i], "--check-def-suffix")) {
	    i++;
	    args.check_def_suffix = argv [i];
	} else if (!strcmp (argv [i], "--conservative")) {
	    args.conservative = 1;
	} else {
	    if (strcmp (argv [i], "--help"))
		fprintf (stderr, "unknown option: %s\n", argv [i]);
	    print_usage ();
	    exit (1);
	}
    }
}

static void
output_xml_begin (char *file)
{
    fprintf (yyout, "@@preamble %s.xml %s%s\n",
	     args.db_file, absolute_path, args.lex_suffix);
    fprintf (yyout, 
	     "@\"file id='F%d_%d' file='%s' path='%s' filename_ref='F%d'\"\n",
	     idnum, id_counter++, file, absolute_path, idnum);
}

static void
output_xml_end (void)
{
    fprintf (yyout, "@\"/file\"\n");
}

static void
output_params (struct CDLL *head)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next)
	fprintf (yyout, "@\"param name='%s'/\"\n", p->param_name);
}

static void
output_params2 (struct CDLL *head)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next)
	fprintf (yyout, "@\"param name='%s'\"%s@\"/param\"",
		 p->param_name, p->param_name);
}

static void
output_error_text (int idnum, int first_line, int first_column,
		   int last_line, int last_column)
{
    int c, line = 1, column = 0;
    FILE *fp = fopen (absolute_path, "r");
    while ((c = getc (fp)) != EOF) {
	if ((first_line < line && line < last_line)
	    || (line == first_line && column >= first_column)
	    || (line == last_line  && column < last_column)) {
	    putc (c, yyout);
	}
	if (c == '\n') {
	    line++;
	    column = 0;
	} else {
	    column++;
	}
    }
    fclose (fp);
}

static char *
get_text (int from, int to)
{
    return chop_nl (substring (directive_buf, from - 1, to - 1));
}

static void
output_check_def_macro (char *macro_name, int is_defined)
{
    char *name = estrdup (macro_name);
    sscanf (macro_name, " %s", name);
    if (is_defined)
	fprintf (yyout, "#define %s%s 1\n", name, args.check_def_suffix);
    else
	fprintf (yyout, "#undef  %s%s\n", name, args.check_def_suffix);
    EFREE (name);
}

static void
tadd (char *file)
{
    char *path, *lex, *ret;

    ret = realpath (file, absolute_path);
    if (ret == NULL) {
	perror ("realpath"); exit (1);
    }
    pi_add (file, absolute_path);
    idnum = pi_path2idnum (absolute_path);

    yyin = fopen (file, "r");
    if (yyin == NULL) {
	perror ("fopen"); return;
    }

    easprintf (&path, "%s/%s",  args.dir, absolute_path);
    easprintf (&lex, "%s/%s%s", args.dir, absolute_path, args.lex_suffix);

    yyout   = fopen_with_mkdir (path, "w");
    if (yyout == NULL) {
	perror ("fopen"); exit (1);
    }

    lexout = fopen_with_mkdir (lex, "w");
    if (lexout == NULL) {
	perror ("fopen"); exit (1);
    }

    yylloc.first_line   = yylloc.last_line   = 1;
    yylloc.first_column = yylloc.last_column = 0;
    
    output_xml_begin (file);
    yyparse ();
    output_xml_end   ();

    fclose (yyin); fclose (yyout); fclose (lexout);
    EFREE (path); EFREE (lex);
}

/* ==================================================================== */
int
main (int argc, char *argv [])
{
    int i;
    char *db_path, *dir;

#if defined(USE_TEF)
    TEF_is_in_main = 1;
#endif

    dir = getenv ("TBCPPA_DIR");
    if (dir != NULL)
	args.dir = dir;

    process_options (argc, argv);

    easprintf (&db_path, "%s/%s", args.dir, args.db_file);
    pi_init (db_path);
    EFREE (db_path);
    absolute_path = emalloc (path_max + 1); /* this must be after pi_init() */

    for (i = 0; i < args.file_num; i++) {
	tadd (args.file [i]);
    }

    pi_final ();
    EFREE (absolute_path);
    return 0;
}

#if 0
  The content of MOVE_FMT2 "%s %s" is correct?
  				     
  How about adding NEWLNIE_opt to func_macro_call?
  => It causes reduce/reduce conflict
  How about using GLR, which saves '#define FOO BAR ('
  => %glr-parser does not work with lexical tie-in.
#endif
